import numpy as np
from .utils import precision, recall, Fmeasure

##################################################
### FRAMEWISE METRICS
##################################################

def TP(data,target):
    return np.sum(np.logical_and(data == 1, target == 1))

def FP(data,target):
    return np.sum(np.logical_and(data == 1, target == 0))

def FN(data,target):
    return np.sum(np.logical_and(data == 0, target == 1))




def framewise(output,target):
    tp,fp,fn = TP(output, target),FP(output, target),FN(output, target),

    P_f,R_f,F_f = precision(tp,fp), recall(tp,fn), Fmeasure(tp,fp,fn)
    return P_f,R_f,F_f


##################################################
### NOTEWISE METRICS
##################################################


def notewise(match,est_pitches,ref_pitches):
    prec = float(len(match))/(len(est_pitches)+np.finfo(float).eps)
    rec = float(len(match))/(len(ref_pitches)+np.finfo(float).eps)
    F = 2*prec*rec/(prec+rec+np.finfo(float).eps)
    return prec, rec, F



import numpy as np
from .utils import precision, recall, Fmeasure, make_note_index_matrix, even_up_rolls, get_loudness

########################################
### Loudness
########################################

# TESTED
def false_negative_loudness(match, vel_target, intervals_target, dt=1):
    # in moving average, include the notes whose onsets are in the range of [t-dt, t+dt)

    if len(match) == 0:
        return 0.0
    else:
        matched_targets, matched_outputs = zip(*match)
        unmatched_targets= list(set(range(len(vel_target)))-set(matched_targets))

        if len(unmatched_targets) == 0:
            return 0.0
        else:
            # unmatched_vels = vel_target[unmatched_targets]
            # avg_unmatched = np.mean(unmatched_vels)
            unmatched_intervals = intervals_target[unmatched_targets]
            unmatched_vels_normed = []
            for unmatched_idx in unmatched_targets:
                recent_vels = []
                for idx in range(len(intervals_target)):
                    if intervals_target[idx][0] >= intervals_target[unmatched_idx][0] - dt and intervals_target[idx][0] < intervals_target[unmatched_idx][0] + dt:
                        recent_vels.append(vel_target[idx])
                unmatched_vels_normed.append(float(vel_target[unmatched_idx]) / np.mean(recent_vels))

            return np.mean(unmatched_vels_normed)


# TESTED
def loudness_ratio_false_negative(notes_target, intervals_target, vel_target, match, min_dur=0.05):
    # loudness ratio of false negative
    # ratio = FN velocity / max loudness in ground truth at onset

    if len(match) == 0:
        return 0.0

    # compute false negatives
    matched_targets, matched_outputs = zip(*match)
    fn_indexs = [idx for idx in range(len(notes_target)) if idx not in matched_targets]
    if len(fn_indexs) == 0:
        return 0.0

    ratios = []
    for i in fn_indexs:
        # loop over false negatives (missing notes)
        note_target = notes_target[i]
        onset_target = intervals_target[i][0]
        velocity = vel_target[i]

        ccn_idxes = [idx for idx in range(len(notes_target)) if intervals_target[idx][0] <= onset_target and intervals_target[idx][1] >= onset_target]
        concurrent_notes_loudnesses = [get_loudness(notes_target[ccn_idxes[idx]], vel_target[ccn_idxes[idx]], onset_target - intervals_target[ccn_idxes[idx]][0]) for idx in range(len(ccn_idxes))]
        ratio = velocity / max(concurrent_notes_loudnesses)
        ratios.append(ratio)

    return np.mean(ratios)



import numpy as np
import os
import pretty_midi as pm
import copy


def make_roll(intervals,pitches,shape,fs=100):
    assert len(intervals) == len(pitches)

    roll = np.zeros(shape)
    for pitch, (start,end) in zip(pitches,intervals):
        # use int() instead of int(round()) to be coherent with PrettyMIDI.get_piano_roll()
        start = int(start*fs)
        end = int(end*fs)
        if start == end:
            end = start+1
        roll[int(pitch),start:end]=1
    return roll


def get_roll_from_times(midi_data,times):
    # quant_piano_roll = midi_data.get_piano_roll(fs=500,times=times)
    # quant_piano_roll = (quant_piano_roll>=7).astype(int)
    roll = np.zeros([128,len(times)])

    for instr in midi_data.instruments:
        for note in instr.notes:
            start = np.argmin(np.abs(times-note.start))
            end = np.argmin(np.abs(times-note.end))
            if start == end:
                end = start+1
            roll[note.pitch,start:end]=1
    return roll


def even_up_rolls(rolls,pad_value=0):
    #Makes roll1 and roll2 of same size.
    lens = [roll.shape[1] for roll in rolls]
    max_len = max(lens)
    output = []
    for roll in rolls:
        if roll.shape[1] < max_len:
            roll = np.concatenate([roll,pad_value*np.ones([roll.shape[0],max_len-roll.shape[1]])],axis=1)
        output += [roll]

    return output

def get_notes_intervals(midi_data,with_vel=False):
    notes= []
    intervals = []
    if with_vel:
        vels = []

    for instr in midi_data.instruments:
        for note in instr.notes:
            notes += [note.pitch]
            intervals += [[note.start,note.end]]
            if with_vel:
                vels += [note.velocity]
    output = [np.array(notes), np.array(intervals)]
    if with_vel:
        output += [np.array(vels)]

    return output

def make_note_index_matrix(notes,intervals,fs=100):
    end_time = np.max(intervals[:,1])
    # Allocate a matrix of zeros - we will add in as we go
    matrix = -np.ones((128, int(fs*end_time)))
    # Make a piano-roll-like matrix holding the index of each note.
    # -1 indicates no note
    for i,(note,interval) in enumerate(zip(notes,intervals)):
        matrix[note,int(interval[0]*fs):int(interval[1]*fs)] = i
    return matrix

def precision(tp,fp):
    #Compute precision for  one file
    pre = tp/(tp+fp+np.finfo(float).eps)

    return pre

def recall(tp,fn):
    #Compute recall for  one file
    rec = tp/(tp+fn+np.finfo(float).eps)
    return rec


def accuracy(tp,fp,fn):
    #Compute accuracy for one file
    acc = tp/(tp+fp+fn+np.finfo(float).eps)
    return acc

def Fmeasure(tp,fp,fn):
    #Compute F-measure  one file
    prec = precision(tp,fp)
    rec = recall(tp,fn)
    return 2*prec*rec/(prec+rec+np.finfo(float).eps)

def get_loudness(midi_pitch, velocity, time):
    # compute decay_rate according to midipitch and note velocity
    time = min(time, 1)
    decay_rate = 0.050532 + 0.021292 * midi_pitch
    loudness = velocity * np.exp(-1.0 * decay_rate * time)
    return loudness



def create_folder(folder):
    if not os.path.exists(folder):
        try:
            os.makedirs(folder)
        except:
            print("create folder error")


def get_time(data,bar,beat,sub_beat):

    first_sig = data.time_signature_changes[0]
    if first_sig.numerator == 1 and first_sig.denominator == 4:
        bar += 1

    PPQ = data.resolution
    downbeats = data.get_downbeats()
    try:
        bar_t = downbeats[bar]
    except IndexError as e:
        if bar == len(downbeats):
            # Instead of the first beat of the bar after the last one (that doesn't exist),
            # We take the before-last beat of the last bar
            bar_t = downbeats[-1]
            beat = 'last'
        else:
            raise e


    time_sigs = data.time_signature_changes
    last_sig = True
    for i, sig in enumerate(time_sigs):
        if sig.time > bar_t:
            last_sig = False
            break

    if last_sig:
        current_sig = time_sigs[i]
    else:
        current_sig = time_sigs[i-1]

    if beat == 'last':
        beat = current_sig.numerator - 1

    try:
        assert beat < current_sig.numerator
    except AssertionError:
        print(downbeats)
        for sig in time_sigs:
            print(sig)
        print('-----------')
        print(bar,beat, bar_t)
        print(current_sig)
        raise AssertionError

    beat_ticks = PPQ * 4 / current_sig.denominator
    tick = data.time_to_tick(bar_t) + beat * beat_ticks + sub_beat*beat_ticks/2
    if tick != int(tick):
        print(bar,beat,sub_beat)
        print(current_sig)
        print(tick)
        raise TypeError('Tick is not int!!!')
    else:
        tick = int(tick)
    time =data.tick_to_time(tick)

    return time


def str_to_bar_beat(string):
    if '.' in string:
        str_split = string.split('.')
        bar = int(str_split[0])
        beat = int(str_split[1])
        output = [bar,beat,0]

        if len(str_split)>2:
            sub_beat=int(str_split[2])
            output[2] = sub_beat
    else:
        output = [int(string),0,0]
    return output

def apply_sustain_control_changes(midi):
    all_CCs = []
    for instr in midi.instruments:
        all_CCs += instr.control_changes
    all_pedals = [cc for cc in all_CCs if cc.number==64]
    pedals_sorted = sorted(all_pedals,key=lambda x: x.time)

    #Add an extra pedal off at the end, just in case
    pedals_sorted += [pm.ControlChange(64,0,midi.get_end_time())]

    #Create a pedal_ON array such that pedal_ON[i]>0 iff pedal is on at tick i
    #If pedal_ON[i]>0, its value is the time at which pedal becomes off again
    pedal_ON = np.zeros(midi._PrettyMIDI__tick_to_time.shape,dtype=float)
    # -1 if pedal is currently off, otherwise tick time of first time it is on.
    ON_idx = -1
    for cc in pedals_sorted:
        if cc.value > 64:
            if ON_idx < 0:
                ON_idx = midi.time_to_tick(cc.time)
            else:
                # Pedal is already ON
                pass
        else:
            if ON_idx>0:
                pedal_ON[ON_idx:midi.time_to_tick(cc.time)]=cc.time
                ON_idx = -1
            else:
                # Pedal is already OFF
                pass

    # Copy to keep time signatures and tempo changes, but remove notes and CCs
    new_midi = copy.deepcopy(midi)
    new_midi.instruments = []


    # Store the notes per pitch, to trim them afterwards.
    all_notes = np.empty([128],dtype=object)
    for i in range(128):
        all_notes[i] = []


    for instr in midi.instruments:

        # First, extend all the notes until the pedal is off
        for note in instr.notes:
            start_tick = midi.time_to_tick(note.start)
            end_tick = midi.time_to_tick(note.end)

            if np.any(pedal_ON[start_tick:end_tick]>0):
                # Pedal is on while note is on
                end_pedal = np.max(pedal_ON[start_tick:end_tick])
                note.end = max(note.end,end_pedal)
            else:
                # Pedal is not on while note is on, no modifications needed
                pass
            all_notes[note.pitch] += [note]

    new_instr = pm.Instrument(program=pm.instrument_name_to_program('Acoustic Grand Piano'),name="Piano")

    # Then, trim notes so they don't overlap
    for note_list in all_notes:
        if note_list != []:
            note_list = sorted(note_list,key=lambda x: x.start)
            for note_1,note_2 in zip(note_list[:-1],note_list[1:]):
                note_1.end = min(note_1.end,note_2.start)
                new_instr.notes += [note_1]
            new_instr.notes += [note_list[-1]]

    new_midi.instruments.append(new_instr)

    return new_midi


def import_features(results,features_to_use):

    all_feat = []
    for feat in features_to_use:
        if feat == 'valid_cons':
            value_hut78 = results["cons_hut78_output"]
            value_har18 = results["cons_har18_output"]
            value_har19 = results["cons_har19_output"]

            value = value_hut78[:-1]+value_har18[0:2]+value_har18[3:]+value_har19[:-2]

        else:
            value = results[feat]
        if type(value) is tuple:
            all_feat += list(value)
        elif type(value) is list:
            all_feat += value
        elif type(value) is float:
            all_feat += [np.float64(value)]
        else:
            all_feat += [value]

    all_feat = [float(elt) for elt in all_feat]


    return all_feat



import numpy as np


########################################
### Polyphony --- discarded
########################################

def polyphony_level_diff(roll_output,roll_target):
    poly_output = np.sum(roll_output,axis=0)
    poly_target = np.sum(roll_target,axis=0)

    poly_diff = np.abs(poly_output-poly_target)

    return np.mean(poly_diff),np.std(poly_diff),np.min(poly_diff),np.max(poly_diff)


# discarded
def false_negative_polyphony_level(roll_target,intervals_target,match):
    fs = 100

    if len(match) == 0:
        unmatched_targets = list(range(intervals_target))
    else:
        matched_targets, matched_outputs = zip(*match)
        # unmatched_targets= list(set(range(len(vel_target)))-set(matched_targets))
        unmatched_targets= list(set(range(len(intervals_target)))-set(matched_targets))

    unmatched_intervals = intervals_target[unmatched_targets,:]

    all_avg_poly = []

    for [start,end] in unmatched_intervals:
        start_idx = int(round(start*fs))
        end_idx = int(round(end*fs))
        avg_poly = np.mean(np.sum(roll_target[:,start_idx:end_idx],axis=0))
        all_avg_poly += [avg_poly]

    return all_avg_poly



import numpy as np
from .utils import *


##############################################
#### Framewise
##############################################


# TESTED
def specific_pitch_framewise(output,target,fs,n_semitones,down_only=False,delta=0.05):

    FPs = np.logical_and(output == 1, target == 0)

    target_shift_down = np.concatenate([target[n_semitones:,:],np.zeros([n_semitones,target.shape[1]])],axis=0)
    target_shift_up = np.concatenate([np.zeros([n_semitones,target.shape[1]]),target[:-n_semitones,:]],axis=0)

    match_down = FPs*target_shift_up # correspond to when an output is matched with a target n_semitones below
    match_up = FPs*target_shift_down # correspond to when an output is matched with a target n_semitones above

    # import matplotlib.pyplot as plt
    # plt.subplot(311)
    # plt.imshow(output)
    # plt.subplot(312)
    # plt.imshow(match_down)
    # plt.subplot(313)
    # plt.imshow(match_up)
    # plt.show()

    delta_steps = int(round(delta*fs))
    delta_steps = min(delta_steps, target.shape[1])  # limit delta_steps within segment length
    continuation_mask = np.concatenate([np.concatenate([target[:,i:],np.zeros([target.shape[0],i])],axis=1)[:,:,None] for i in range(delta_steps)],axis=2)

    match_past = FPs*np.all(continuation_mask==0,axis=2).astype(int)

    if down_only:
        n_match = np.sum(match_down*match_past)
    else:
        match_pitch = np.logical_or(match_down,match_down).astype(int)
        n_match = np.sum(match_pitch*match_past)

    if np.sum(FPs) == 0:
        return 0.0,0.0
    else:
        n_match = float(n_match)
        n_FP = np.sum(FPs)
        n_tot = np.sum(output)

        return n_match/n_FP, n_match/n_tot


########################################
### Notewise
########################################

# TESTED
def specific_pitch_notewise(notes_output, intervals_output, notes_target, intervals_target, match, n_semitones, down_only=False, ratio=0.8):
    # return two ratios:
    # 1. the proportion of specific pitch mistakes among false positives
    # 2. the proportion of specific pitch mistakes among all detected notes
    if len(match) == 0:
        return 0.0, 0.0

    # get false positives
    matched_targets, matched_outputs = zip(*match)
    fp_idxs = [idx for idx in range(notes_output.shape[0]) if idx not in matched_outputs]
    if len(fp_idxs) == 0:
        return 0.0, 0.0

    n_specific_pitch = 0.0
    for i in fp_idxs:
        # loop over all false positives
        note_output = notes_output[i]
        interval_output = intervals_output[i]
        is_specific_pitch_error = False

        note_down = note_output - n_semitones
        intervals_down = [intervals_target[idx] for idx in range(len(notes_target)) if notes_target[idx] == note_down]

        for interval in intervals_down:
            if (min(interval_output[1], interval[1]) - max(interval_output[0], interval[0])) / (interval_output[1] - interval_output[0]) > ratio:
                is_specific_pitch_error = True
                break

        if not down_only and not is_specific_pitch_error:
            note_up = note_output + n_semitones
            intervals_up = [intervals_target[idx] for idx in range(len(notes_target)) if notes_target[idx] == note_up]

            for interval in intervals_up:
                if (min(interval_output[1], interval[1]) - max(interval_output[0], interval[0])) / (interval_output[1] - interval_output[0]) > ratio:
                    is_specific_pitch_error = True
                    break

        if is_specific_pitch_error:
            n_specific_pitch += 1.0

    return n_specific_pitch / len(fp_idxs), n_specific_pitch / len(notes_output)



import numpy as np
from .utils import precision, recall, Fmeasure, make_note_index_matrix, even_up_rolls


##############################################
#### Framewise highest and lowest voice
##############################################

def get_highest(roll):
    # when no note, returns -1
    highest = np.argmax(roll[::-1,:],axis=0)
    highest[highest!=0] = roll.shape[0]-1-highest[highest!=0]
    highest[highest==0] = -1
    return highest

def get_lowest(roll):
    # when no note, returns roll.shape[0]
    # we give value roll.shape[0] rather than roll.shape[0]-1, otherwise,
    # we cannot distinguish the case where there is no note and where the lowest note is roll.shape[0]-1
    lowest = np.argmax(roll,axis=0)
    # We must make the distinction between no note, and lowest note is 0
    lowest[np.logical_and(lowest==0,roll[0,:]==0)] = roll.shape[0]
    return lowest

# TESTED
def framewise_highest(output, target):

    highest = get_highest(target)

    # plot_piano_roll(target)
    # print(highest)

    highest_nonzero = highest[highest!=-1]
    frames_nonzero = np.arange(len(highest))[highest!=-1]

    tp = np.sum(output[highest_nonzero,frames_nonzero])

    fn = np.sum(output[highest_nonzero,frames_nonzero]==0)

    i,j = np.indices(target.shape)
    mask = [i>highest]
    ### Count all false positives above highest reference note)
    fp = np.sum(output[tuple(mask)])

    return precision(tp,fp),recall(tp, fn), Fmeasure(tp,fp,fn)


# TESTED
def framewise_lowest(output, target):

    lowest = get_lowest(target)

    # We give value target.shape[0] when there are no notes by convention
    lowest_nonzero = lowest[lowest!=target.shape[0]]
    frames_nonzero = np.arange(len(lowest))[lowest!=target.shape[0]]

    tp = np.sum(output[lowest_nonzero,frames_nonzero])

    fn = np.sum(output[lowest_nonzero,frames_nonzero]==0)



    i,j = np.indices(target.shape)
    mask = [i<lowest]
    ### Count all false positives above highest reference note)
    fp = np.sum(output[tuple(mask)])

    return precision(tp,fp),recall(tp, fn), Fmeasure(tp,fp,fn)

########################################
### Notewise highest and lowest voice
########################################


# TESTED
def notewise_highest(notes_output,intervals_output,notes_target,intervals_target,match,min_dur=0.05):
    #min_dur represents the minimum duration a note has to be the highest to be considered
    #in the skyline
    if len(match) == 0:
        return 0.0, 0.0, 0.0
    else:

        fs = 100

        # Get the list of highest notes
        target_refs = make_note_index_matrix(notes_target,intervals_target)
        output_refs = make_note_index_matrix(notes_output,intervals_output)
        target_refs,output_refs = even_up_rolls([target_refs,output_refs],pad_value=-1)

        roll_target = (target_refs!=-1).astype(int)
        roll_output = (output_refs!=-1).astype(int)

        highest = get_highest(roll_target)

        # print(highest)
        # plt.imshow(roll_target)
        # plt.show()

        highest_nonzero = highest[highest!=-1]
        frames_nonzero = np.arange(len(highest))[highest!=-1]

        highest_notes_idx, count = np.unique(target_refs[highest_nonzero,frames_nonzero],return_counts=True)
        highest_notes_idx = highest_notes_idx[count/float(fs) > min_dur]

        # Compute true positives
        # NB: matching gives indexes (idx_target,idx_output)


        matched_targets, matched_outputs = zip(*match)
        matched_targets_is_highest = [idx for idx in matched_targets if idx in highest_notes_idx]
        tp = len(matched_targets_is_highest)

        # Compute false negatives
        unmatched_targets= list(set(range(len(notes_target)))-set(matched_targets))
        unmatched_targets_is_highest = [idx for idx in unmatched_targets if idx in highest_notes_idx]
        fn = len(unmatched_targets_is_highest)

        # Compute false positives
        # Count all false positives that are above the highest note
        i,j = np.indices(target_refs.shape)
        higher_mask = [i>highest]
        higher_notes_idx, count = np.unique(output_refs[tuple(higher_mask)],return_counts=True)
        count = count[higher_notes_idx!= -1]
        higher_notes_idx = higher_notes_idx[higher_notes_idx!= -1]
        higher_notes_idx = higher_notes_idx[count/float(fs) > min_dur]

        unmatched_outputs= list(set(range(len(notes_output)))-set(matched_outputs))
        unmatched_outputs_is_higher = [idx for idx in unmatched_outputs if idx in higher_notes_idx]
        fp = len(unmatched_outputs_is_higher)

        # fig,((ax0,ax1),(ax2,ax3)) = plt.subplots(2,2)
        # ax0.imshow(roll_target,aspect='auto',origin='lower')
        # ax1.imshow(roll_output,aspect='auto',origin='lower')
        # display1 = np.zeros_like(roll_output)
        # display2 = np.zeros_like(roll_output)
        # for i in matchighest_lowest_voicehed_targets:
        #     display1[notes_target[i],int(intervals_target[i][0]*fs):int(intervals_target[i,1]*fs)] = 1
        # for i in unmatched_outputs_is_higher:
        #     display2[notes_output[i],int(intervals_output[i][0]*fs):int(intervals_output[i,1]*fs)] = 1

        # display2[higher_mask]+=3

        # ax2.imshow(display1,aspect='auto',origin='lower')
        # ax3.imshow(display2,aspect='auto',origin='lower')
        # plt.show()

        return precision(tp,fp),recall(tp, fn), Fmeasure(tp,fp,fn)


# TESTED
def notewise_lowest(notes_output,intervals_output,notes_target,intervals_target,match,min_dur=0.05):

    if len(match) == 0:
        return 0.0,0.0,0.0
    else:
        #min_dur represents the minimum duration a note has to be the highest to be considered
        #in the skyline
        fs = 100

        # Get the list of highest notes
        target_refs = make_note_index_matrix(notes_target,intervals_target)
        output_refs = make_note_index_matrix(notes_output,intervals_output)
        target_refs,output_refs = even_up_rolls([target_refs,output_refs],pad_value=-1)

        roll_target = (target_refs!=-1).astype(int)
        roll_output = (output_refs!=-1).astype(int)

        lowest = get_lowest(roll_target)
        lowest_nonzero = lowest[lowest!=roll_target.shape[0]]
        frames_nonzero = np.arange(len(lowest))[lowest!=roll_target.shape[0]]

        lowest_notes_idx, count = np.unique(target_refs[lowest_nonzero,frames_nonzero],return_counts=True)
        lowest_notes_idx = lowest_notes_idx[count/float(fs) > min_dur]

        # Compute true positives
        # NB: matching gives indexes (idx_target,idx_output)

        matched_targets, matched_outputs = zip(*match)
        matched_targets_is_lowest = [idx for idx in matched_targets if idx in lowest_notes_idx]
        tp = len(matched_targets_is_lowest)

        # Compute false negatives
        unmatched_targets= list(set(range(len(notes_target)))-set(matched_targets))
        unmatched_targets_is_lowest = [idx for idx in unmatched_targets if idx in lowest_notes_idx]
        fn = len(unmatched_targets_is_lowest)

        # Compute false positives
        # Count all false positives that are above the lowest note
        i,j = np.indices(target_refs.shape)
        lower_mask = [i<lowest]
        lower_notes_idx, count = np.unique(output_refs[tuple(lower_mask)],return_counts=True)
        count = count[lower_notes_idx!= -1]
        lower_notes_idx = lower_notes_idx[lower_notes_idx!= -1]
        # print(lower_notes_idx, count)
        lower_notes_idx = lower_notes_idx[count/float(fs) > min_dur]

        unmatched_outputs= list(set(range(len(notes_output)))-set(matched_outputs))
        unmatched_outputs_is_lower = [idx for idx in unmatched_outputs if idx in lower_notes_idx]
        fp = len(unmatched_outputs_is_lower)

        # import matplotlib.pyplot as plt
        # fig,((ax0,ax1),(ax2,ax3)) = plt.subplots(2,2)
        # ax0.imshow(roll_target,aspect='auto',origin='lower')
        # ax1.imshow(roll_output,aspect='auto',origin='lower')
        # display1 = np.zeros_like(roll_output)
        # display2 = np.zeros_like(roll_output)
        # for i in matched_targets:
        #     display1[notes_target[i],int(intervals_target[i][0]*fs):int(intervals_target[i,1]*fs)] = 1
        # for i in unmatched_outputs_is_lower:
        #     display2[notes_output[i],int(intervals_output[i][0]*fs):int(intervals_output[i,1]*fs)] = 1
        #
        # display2[lower_mask]+=3
        #
        # ax2.imshow(display1,aspect='auto',origin='lower')
        # ax3.imshow(display2,aspect='auto',origin='lower')
        # plt.show()

        return precision(tp,fp),recall(tp, fn), Fmeasure(tp,fp,fn)


# TESTED
def correct_highest_lowest_note_framewise(output, target):
    # return two parameters, the proportion of correct highest notes and lowest notes framewise
    highest_output = get_highest(output)
    highest_target = get_highest(target)
    correct_highest_seq = [int(highest_output[idx] == highest_target[idx]) for idx in range(len(highest_output))]
    lowest_output = get_lowest(output)
    lowest_target = get_lowest(target)
    correct_lowest_seq = [int(lowest_output[idx] == lowest_target[idx]) for idx in range(len(lowest_output))]

    correct_highest_count = correct_highest_seq.count(1)
    correct_lowest_count = correct_lowest_seq.count(1)

    return float(correct_highest_count) / len(correct_highest_seq), float(correct_lowest_count) / len(correct_lowest_seq)



import numpy as np
from numpy import random

SMALL_VALUE = 0.00001

def calculate_stds(data, means):
	# calculate stds for k-means result
	data_groups = [[] for i in range(len(means))]
	for ioi in data:
		data_groups[np.argmin(abs(np.array(means) - ioi))].append(ioi)
	stds = []
	for group in data_groups:
		if len(group) <= 1:
			stds.append(0.0)
		else:
			stds.append(np.std(group, ddof=1))
	return stds


# TESTED
def rhythm_histogram(intervals_output, intervals_target):
    # return the logged spectral flatness ratio of the IOIs, where spectral flatness ratio is defined as the ratio of the geometric mean of the histogram over its arithmetic mean.
    # 1. spectral flatness for IOI of the output transcription
    # 2. spectral flatness for IOI of the ground truth music piece

    # order note intervals by onsets
    onsets_output = [interval[0] for interval in intervals_output]
    onsets_target = [interval[0] for interval in intervals_target]
    onsets_output.sort()
    onsets_target.sort()

    ioi_output = [onsets_output[idx+1] - onsets_output[idx] for idx in range(len(onsets_output)-1)]
    ioi_target = [onsets_target[idx+1] - onsets_target[idx] for idx in range(len(onsets_target)-1)]

    # generate bins
    bins = [i*0.01 for i in range(10)]
    bins += [0.1+i*0.1 for i in range(20)]

    histogram_output = np.histogram(ioi_output, bins=bins)[0]
    histogram_target = np.histogram(ioi_target, bins=bins)[0]

    histogram_output = [max(SMALL_VALUE, histogram_output[idx]) for idx in range(len(histogram_output))]
    histogram_target = [max(SMALL_VALUE, histogram_target[idx]) for idx in range(len(histogram_target))]

    # logged geometric means
    log_gmean_output = np.mean(np.log(histogram_output))
    log_gmean_target = np.mean(np.log(histogram_target))
    # arithmetic means
    mean_output = np.mean(histogram_output)
    mean_target = np.mean(histogram_target)
    # logged spectral flatness
    log_spectral_flatness_output = log_gmean_output - np.log(mean_output)
    log_spectral_flatness_target = log_gmean_target - np.log(mean_target)

    # print(log_spectral_flatness_output)
    # print(log_spectral_flatness_target)
    # print(log_spectral_flatness_output / log_spectral_flatness_target)
    # import matplotlib.pyplot as plt
    # plt.subplot(211)
    # plt.bar(np.arange(len(histogram_target)), histogram_target)
    # plt.subplot(212)
    # plt.bar(np.arange(len(histogram_output)), histogram_output)
    # plt.show()

    return log_spectral_flatness_output, log_spectral_flatness_output-log_spectral_flatness_target


# TESTED
def rhythm_dispersion(intervals_output, intervals_target):
    # return changes in k-means clusters
    # 1. change in standard deviations
    # 2. center drift

    # order note intervals by onsets
    onsets_output = [interval[0] for interval in intervals_output]
    onsets_target = [interval[0] for interval in intervals_target]
    onsets_output.sort()
    onsets_target.sort()


    ioi_output = [onsets_output[idx+1] - onsets_output[idx] for idx in range(len(onsets_output)-1)]
    ioi_target = [onsets_target[idx+1] - onsets_target[idx] for idx in range(len(onsets_target)-1)]

    # initialise cluster
    bins = [i*0.02 for i in range(5)]
    bins += [0.1+i*0.2 for i in range(10)]
    histogram_target = np.histogram(ioi_target, bins=bins)[0]
    means = []
    for i in range(len(histogram_target)):
        if histogram_target[i] > 0 and (i == 0 or histogram_target[i] > histogram_target[i-1]) and (i == len(histogram_target)-1 or histogram_target[i] >= histogram_target[i+1]):
            means.append(np.mean([bins[i], bins[i+1]]))

    if len(means) == 0:
        return 0.0, 0.0

    # k-means on target intervals
    moving = 1
    while moving > 0.0001:
        # cluster
        ioi_target_labels = [np.argmin(abs(np.array(means) - ioi_target[idx])) for idx in range(len(ioi_target))]
        # calculate new means
        new_means = []
        for label in range(len(means)):
            indexs = np.array([i for i in range(len(ioi_target)) if ioi_target_labels[i] == label])
            if len(indexs) > 0:
                # only update cluster if there are some points within it, else discard it.
                new_means.append(np.mean(np.array(ioi_target)[indexs]))
            else:
                means.pop(label)
        # calculate centre moving
        moving = np.sum(abs(np.array(new_means) - np.array(means)))
        # update means
        means = new_means
        # print(means)

    # calculate target std
    stds_target = calculate_stds(ioi_target, means)

    # print('--')
    # k-means on output intervals
    means_output = [means[i] for i in range(len(means))] # copy target means
    moving = 1
    while moving > 0.0001:
        # cluster
        ioi_output_labels = [np.argmin(abs(np.array(means_output) - ioi_output[idx])) for idx in range(len(ioi_output))]
        # calculate new means
        new_means_output = []
        for label in range(len(means_output)):
            indexs = np.array([i for i in range(len(ioi_output)) if ioi_output_labels[i] == label])
            if len(indexs) > 0:
                new_means_output.append(np.mean(np.array(ioi_output)[indexs]))
            else:
                # if there is no points within this cluster, keep the initial cluster mean
                new_means_output.append(means_output[label])
        # calculate centre moving
        moving = np.sum(abs(np.array(new_means_output) - np.array(means_output)))
        # update means
        means_output = new_means_output
        # print(means_output)

    # calculate stds for output
    stds_output = calculate_stds(ioi_output, means_output)
    # calcuate std chage
    stds_change = list(np.array(stds_output) - np.array(stds_target))

    # print(stds_target)
    # print(stds_output)

    # cluster centre drift
    drifts = [abs(means[idx] - means_output[idx]) for idx in range(len(means))]

    # # test with graphs
    # histogram_output = np.histogram(ioi_output, bins=bins)[0]
    # histogram_output = [max(0, histogram_output[idx]) for idx in range(len(histogram_output))]
    # histogram_target = [max(0, histogram_target[idx]) for idx in range(len(histogram_target))]
    # import matplotlib.pyplot as plt
    # plt.subplot(411)
    # plt.bar(np.arange(len(histogram_target)), histogram_target)
    # plt.subplot(412)
    # plt.bar(np.arange(len(histogram_output)), histogram_output)
    # plt.subplot(413)
    # plt.bar(np.arange(len(stds_change)), stds_change)
    # plt.subplot(414)
    # plt.bar(np.arange(len(drifts)), drifts)
    # plt.show()

    return [np.mean(stds_change),np.min(stds_change),np.max(stds_change)], [np.mean(drifts),np.min(drifts),np.max(drifts)]



import numpy as np
from .utils import precision, recall, Fmeasure, make_note_index_matrix, even_up_rolls



########################################
### Repeated/merged notes
########################################


def repeated_notes(notes_output, intervals_output, notes_target, intervals_target, match, tol=0.8):
    if len(match) == 0:
        matched_targets, matched_outputs = (), ()
    else:
        matched_targets, matched_outputs = zip(*match)
    unmatched_outputs = list(set(range(len(notes_output))) - set(matched_outputs))

    if len(unmatched_outputs) == 0:
        return 0.0, 0.0

    # point unmatched outputs to a target note
    unmatched_outputs_pointer = [-1] * len(unmatched_outputs)
    for idx in range(len(unmatched_outputs)):
        pitch = notes_output[unmatched_outputs[idx]]
        interval = intervals_output[unmatched_outputs[idx]]
        duration = interval[1] - interval[0]
        # find overlap note in target
        for j in range(len(notes_target)):
            if notes_target[j] == pitch and (min(interval[1], intervals_target[j][1]) - max(interval[0], intervals_target[j][0])) > (0.8 * duration):
                unmatched_outputs_pointer[idx] = j

    pointed_targets = list(filter(lambda x:x != -1, unmatched_outputs_pointer))
    n_repeat = len(pointed_targets) - len(set(pointed_targets))

    return float(n_repeat) / len(unmatched_outputs), float(n_repeat) / len(notes_output)


def merged_notes(notes_output, intervals_output, notes_target, intervals_target, match, tol=0.8):
    if len(match) == 0:
        matched_targets, matched_outputs = (), ()
    else:
        matched_targets, matched_outputs = zip(*match)
    unmatched_targets = list(set(range(len(notes_target))) - set(matched_targets))

    if len(unmatched_targets) == 0:
        return 0.0, 0.0

    # point unmatched targets to an output note
    unmatched_targets_pointer = [-1] * len(unmatched_targets)
    for idx in range(len(unmatched_targets)):
        pitch = notes_target[unmatched_targets[idx]]
        interval = intervals_target[unmatched_targets[idx]]
        duration = interval[1] - interval[0]
        # find overlap note in output
        for j in range(len(notes_output)):
            if notes_output[j] == pitch and (min(interval[1], intervals_output[j][1]) - max(interval[0], intervals_output[j][0])) > (0.8 * duration):
                unmatched_targets_pointer[idx] = j

    pointed_outputs = list(filter(lambda x:x != -1, unmatched_targets_pointer))
    # print(pointed_outputs)
    pointed_outputs_set = set(pointed_outputs)
    n_merged = 0
    for output in pointed_outputs_set:
        if pointed_outputs.count(output) > 1:
            # print(output)
            n_merged += 1

    # print(n_merged)
    return float(n_merged) / len(unmatched_targets), float(n_merged) / len(notes_output)



######################################################
# old definition, count starting notes
######################################################


def repeated_notes_old_stuff(notes_output,intervals_output,notes_target,intervals_target,match,tol=0.8):
    # Here, any note that is a false positive an overlaps with a ground truth note for more
    # than tol percent of its duration is considered a repeated note

    if len(match) == 0:
        return 0.0, 0.0
    else:
        fs = 500

        matched_targets, matched_outputs = zip(*match)
        matched_targets = np.array(matched_targets)
        matched_outputs = list(matched_outputs)
        unmatched_outputs= list(set(range(len(notes_output)))-set(matched_outputs))

        if len(unmatched_outputs)==0:
            # No false positives, return zero
            return 0.0,0.0
        else:

            repeated = []

            target_refs = make_note_index_matrix(notes_target,intervals_target,fs)
            output_refs = make_note_index_matrix(notes_output,intervals_output,fs)
            target_refs,output_refs = even_up_rolls([target_refs,output_refs],pad_value=-1)

            roll_target = (target_refs!=-1).astype(int)
            roll_output = (output_refs!=-1).astype(int)

            if len(unmatched_outputs) == 0:
                return 0.0,0.0
            else:
                for idx in unmatched_outputs:
                    roll_idx = output_refs==idx
                    overlap = np.sum(roll_target[roll_idx])/float(fs)
                    note_duration = intervals_output[idx][1] -  intervals_output[idx][0]
                    if overlap/note_duration > tol:
                        repeated += [idx]

                n_repeat = float(len(repeated))
                tot_err = len(unmatched_outputs)
                tot_notes = len(notes_output)

                return n_repeat/tot_err, n_repeat/tot_notes


# TESTED
def merged_notes_old_stuff(notes_output,intervals_output,notes_target,intervals_target,match,tol=0.8):
    # Here, any note that is a false positive an overlaps with a ground truth note for more
    # than tol percent of its duration is considered a repeated note

    if len(match) == 0:
        return 0.0, 0.0
    else:
        fs = 500

        matched_targets, matched_outputs = zip(*match)
        matched_targets = np.array(matched_targets)
        unmatched_targets= list(set(range(len(notes_target)))-set(matched_targets))

        if len(unmatched_targets)==0:
            # No false positives, return zero
            return 0.0,0.0
        else:

            repeated = []

            target_refs = make_note_index_matrix(notes_target,intervals_target,fs)
            output_refs = make_note_index_matrix(notes_output,intervals_output,fs)
            target_refs,output_refs = even_up_rolls([target_refs,output_refs],pad_value=-1)

            roll_target = (target_refs!=-1).astype(int)
            roll_output = (output_refs!=-1).astype(int)

            for idx in unmatched_targets:
                roll_idx = target_refs==idx
                overlap = np.sum(roll_output[roll_idx])/float(fs)
                note_duration = intervals_target[idx][1] -  intervals_target[idx][0]
                if overlap/note_duration > tol:
                    repeated += [idx]


            n_repeat = float(len(repeated))
            tot_err = len(unmatched_targets)
            tot_notes = len(notes_output)


            return n_repeat/tot_err, n_repeat/tot_notes


##################################
#### OLD STUFF
##################################

def repeated_notes_WITH_CORRECT_ONSET_ONLY(intervals_target,notes_output,intervals_output,match_on):
    # Here, a repeated note is counted if and only if there is a correctly detected note before (onset-only)

    if len(match_on) == 0:
        # No matches, return zero
        return 0.0,0.0
    else:
        matched_targets, matched_outputs = zip(*match_on)
        matched_targets = np.array(matched_targets)
        matched_outputs = list(matched_outputs)
        unmatched_outputs= list(set(range(len(notes_output)))-set(matched_outputs))

        if len(unmatched_outputs)==0:
            # No false positives, return zero
            return 0.0,0.0,[]
        else:

            matched_notes = notes_output[matched_outputs]
            unmatched_notes = notes_output[unmatched_outputs]
            # keys are pitches, values are the target intervals that have been matched to that pitch
            pitch_dict = {}

            print((matched_outputs,unmatched_outputs))
            for note in unmatched_notes:
                if not note in pitch_dict:
                    match_indices = np.where(matched_notes==note)[0]
                    # print match_indices
                    intervals_indices = matched_targets[match_indices]
                    # print note,intervals_indices,intervals_target[intervals_indices]
                    pitch_dict[note] = intervals_target[intervals_indices]
                else:
                    #Only compute that the first time this pitch is encountered (same results)
                    pass

            repeated = []
            for unmatched_idx in unmatched_outputs:
                note = notes_output[unmatched_idx]
                matching_target_intervals = pitch_dict[note]
                # If there are some matched notes with same pitch
                # print matching_target_intervals
                if not matching_target_intervals.size==0:
                    interval_out = intervals_output[unmatched_idx]
                    # print matching_target_intervals, interval_out
                    overlap = (np.minimum(matching_target_intervals[:,1],interval_out[1]) - np.maximum(matching_target_intervals[:,0],interval_out[0]))/(interval_out[1]-interval_out[0])
                    is_repeated = overlap > 0.8
                    print((note, overlap))
                    assert sum(is_repeated.astype(int))==1 or sum(is_repeated.astype(int))==0

                    if np.any(is_repeated):
                        repeated += [unmatched_idx]


            n_repeat = float(len(repeated))
            tot_err = len(unmatched_outputs)
            tot_notes = len(notes_output)

            print((n_repeat, tot_err, tot_notes))

            return n_repeat/tot_err, n_repeat/tot_notes, repeated



import setuptools


setuptools.setup(
    name="peamt", # Replace with your own username
    version="0.1",
    author="Adrien Ycart",
    author_email="adrien.ycart@gmail.com",
    description="PEAMT: a Perceptual Evaluation metric for Automatic Music Transcription",
    long_description="""
    # PEAMT

    This package contains code to run PEAMT, a Perceptual Evaluation metric for Automatic Music Transcription.
    If you use any of this, please cite:

    Adrien Ycart, Lele Liu, Emmanouil Benetos and Marcus Pearce, 2020. ["Investigating the Perceptual Validity of Evaluation Metrics for Automatic Piano Music Transcription"](https://transactions.ismir.net/articles/10.5334/tismir.57), _Transactions of the International Society for Music Information Retrieval (TISMIR)_, 3(1), pp.68–81.

    ```
        @article{ycart2019PEAMT,
           Author = {Ycart, Adrien and Liu, Lele and Benetos, Emmanouil and Pearce, Marcus},
           Booktitle = {Transactions of the International Society for Music Information Retrieval (TISMIR)},
           Title = {Investigating the Perceptual Validity of Evaluation Metrics for Automatic Piano Music Transcription},
           Year = {2020},
           Volume = {3},
           Issue = {1},
           Pages = {68--81},
           DOI = {http://doi.org/10.5334/tismir.57},
        }
    ```

    For more info, please visit: [https://github.com/adrienycart/PEAMT](https://github.com/adrienycart/PEAMT)

    """,
    long_description_content_type="text/markdown",
    url="https://github.com/adrienycart/PEAMT",
    project_urls={
        "Bug Tracker": "https://github.com/adrienycart/PEAMT/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],

    install_requires=[
        'numpy',
        'pretty_midi==0.2.8',
        'mir_eval',
    ],

    package_dir={"": "peamt"},
    packages=setuptools.find_packages(where="peamt"),

    package_data={
        "peamt": ["model_parameters/*.pkl"],
    }
)



import numpy as np

########################################
### Out-of-key notes
########################################


def make_key_mask(target_roll):

    i = np.arange(target_roll.shape[0])
    output = np.zeros([12],dtype=float)
    for p in range(12):
        active_pitch_class = np.max(target_roll[i%12==p,:],axis=0)
        output[p] = np.mean(active_pitch_class)

    return output


# TESTED
def out_key_errors(notes_output,match,mask):

    if len(match) == 0:
        unmatched_outputs = list(range(len(notes_output)))
    else:
        matched_targets, matched_outputs = zip(*match)
        unmatched_outputs= list(set(range(len(notes_output)))-set(matched_outputs))

    if len(unmatched_outputs) == 0:
        return 0.0,0.0
    else:
        unmatched_weights = []
        for i in unmatched_outputs:
            unmatched_weights += [1- mask[notes_output[i]%12]]

        all_weights = []
        for note in notes_output:
            all_weights += [1- mask[note%12]]

        # import matplotlib.pyplot as plt
        # plt.plot(all_weights)
        # plt.show()

        if sum(all_weights)==0:
            return 0.0, 0.0
        else:
            return np.mean(unmatched_weights), sum(unmatched_weights)/sum(all_weights)


# TESTED
def out_key_errors_binary_mask(notes_output,match,mask,mask_thresh=0.1):


    in_mask = mask>mask_thresh

    # import matplotlib.pyplot as plt
    # fig, (ax0,ax1) = plt.subplots(2,1)
    # ax0.plot(mask)
    # ax1.plot(in_mask.astype(int))
    # plt.show()

    if len(match) == 0:
        unmatched_outputs = list(range(len(notes_output)))
    else:
        matched_targets, matched_outputs = zip(*match)
        unmatched_outputs= list(set(range(len(notes_output)))-set(matched_outputs))

    if len(unmatched_outputs) == 0:
        return 0.0,0.0
    else:
        out_key_unmatched = []
        for i in unmatched_outputs:
            if not in_mask[notes_output[i]%12]:
                out_key_unmatched += [notes_output[i]]

        tot_out_key = float(len(out_key_unmatched))
        tot_err = len(unmatched_outputs)
        tot_notes = len(notes_output)

        return tot_out_key/tot_err, tot_out_key/tot_notes
